//Command bitmask
#define SERMASK  0x80  //SerialCommunicationFlag
#define READMASK 0x40  //ReadEnableFlag
#define OPEMASK  0x20  //OutputEnable(LED)
#define TDIMASK  0x10
#define nCSMASK  0x08
#define nCEMASK  0x04
#define TMSMASK  0x02
#define TCKMASK  0x01

//Port Mapping
#define LEDPIN 13	//LED(OutputEnable)
#define TCKPIN 12	//JTAG: TCK
#define TMSPIN 10	//JTAG: TMS
#define TDIPIN  8	//JTAG: TDI
#define TDOPIN 11	//JTAG: TDO
#define nCEPIN  7	//AS  : nCE
#define nCSPIN  6	//AS  : nCS
#define ADOPIN  5	//AS  : DataOut ,PS  :nSTATUS
#define CTSPIN  9	//FT232R_CTS
#define BSYPIN  3	//Busy Flag

//function prototypes
inline void jtag_write(byte dw);
inline byte jtag_rw(byte dw);
inline byte aser_rw(byte dw);
inline void pio_write(byte c);
inline byte pio_read();

//FastIO Macros
#define pinset(pin) {if((pin)<8) bitSet(PORTD,(pin)); else bitSet(PORTB,(pin)-8);}
#define pinclr(pin) {if((pin)<8) bitClear(PORTD,(pin)); else bitClear(PORTB,(pin)-8);}
#define m_digitalWrite(pin,hl) {if((hl)==HIGH) pinset(pin); if((hl)==LOW) pinclr(pin);}
//#define m_digitalWrite digitalWrite
#define m_digitalRead(pin) (((pin)<8?bitRead(PIND,(pin)):bitRead(PINB,(pin)-8))?HIGH:LOW)
//#define m_digitalRead digitalRead

//FastSerialFunctions
void Serial_begin(unsigned long baud);
inline byte Serial_available();
inline byte Serial_read();
inline void Serial_write(byte c);
//#define Serial_begin Serial.begin
//#define Serial_available Serial.available
//#define Serial_read Serial.read
//#define Serial_write Serial.write

//initialize
void setup(){
  Serial_begin(1000000);
  pinMode(TCKPIN,OUTPUT);  pinMode(TMSPIN,OUTPUT);  pinMode(TDIPIN,OUTPUT);
  pinMode(nCEPIN,OUTPUT);  pinMode(nCSPIN,OUTPUT);
  pinMode(LEDPIN,OUTPUT);  pinMode(CTSPIN,OUTPUT);
  pinMode(3,OUTPUT);
  m_digitalWrite(CTSPIN,LOW);
  while(Serial_available()) Serial_read();
}

//main loop
void loop(){
  static byte c, jtag_cnt=0, aser_cnt=0, ret;
  static boolean read_flg, busy;
  ret = Serial_available();
  if(ret>48) m_digitalWrite(CTSPIN,HIGH);  //stop data receiving
  if(ret<32) m_digitalWrite(CTSPIN,LOW);  //restart data receiving
  if(ret){
    busy = true;
    m_digitalWrite(BSYPIN,HIGH); //output busy flag
    c = Serial_read();
    if(jtag_cnt){
      if(read_flg) Serial_write(jtag_rw(c));
      else jtag_write(c);
      jtag_cnt--;
    }else if(aser_cnt){
      if(read_flg) Serial_write(aser_rw(c));
      else jtag_write(c);
      aser_cnt--;
    }else{
      if(c & READMASK) read_flg = true; else read_flg = false;
      if(c & SERMASK){
        m_digitalWrite(TCKPIN,LOW); //bug fix
        if(m_digitalRead(nCSPIN)==HIGH) jtag_cnt = c & 0x3F;
        else  aser_cnt = c & 0x3F;
      }else{
        pio_write(c);
        if(read_flg) Serial_write(pio_read());
      }
    }
  }else{
    if(busy) m_digitalWrite(CTSPIN,HIGH); //improve speed
    busy = false;
    m_digitalWrite(BSYPIN,LOW);  //output busy flag
  }
}

//functions
#define tck() {m_digitalWrite(TCKPIN,HIGH); m_digitalWrite(TCKPIN,LOW);}
inline void jtag_write(byte dw){
  m_digitalWrite(TDIPIN,dw&0x01?HIGH:LOW); tck();
  m_digitalWrite(TDIPIN,dw&0x02?HIGH:LOW); tck();
  m_digitalWrite(TDIPIN,dw&0x04?HIGH:LOW); tck();
  m_digitalWrite(TDIPIN,dw&0x08?HIGH:LOW); tck();
  m_digitalWrite(TDIPIN,dw&0x10?HIGH:LOW); tck();
  m_digitalWrite(TDIPIN,dw&0x20?HIGH:LOW); tck();
  m_digitalWrite(TDIPIN,dw&0x40?HIGH:LOW); tck();
  m_digitalWrite(TDIPIN,dw&0x80?HIGH:LOW); tck();
}
inline byte jtag_rw(byte dw){
  byte dr=0;
  m_digitalWrite(TDIPIN,dw&0x01?HIGH:LOW); if(m_digitalRead(TDOPIN)==HIGH) dr|=0x01; tck();
  m_digitalWrite(TDIPIN,dw&0x02?HIGH:LOW); if(m_digitalRead(TDOPIN)==HIGH) dr|=0x02; tck();
  m_digitalWrite(TDIPIN,dw&0x04?HIGH:LOW); if(m_digitalRead(TDOPIN)==HIGH) dr|=0x04; tck();
  m_digitalWrite(TDIPIN,dw&0x08?HIGH:LOW); if(m_digitalRead(TDOPIN)==HIGH) dr|=0x08; tck();
  m_digitalWrite(TDIPIN,dw&0x10?HIGH:LOW); if(m_digitalRead(TDOPIN)==HIGH) dr|=0x10; tck();
  m_digitalWrite(TDIPIN,dw&0x20?HIGH:LOW); if(m_digitalRead(TDOPIN)==HIGH) dr|=0x20; tck();
  m_digitalWrite(TDIPIN,dw&0x40?HIGH:LOW); if(m_digitalRead(TDOPIN)==HIGH) dr|=0x40; tck();
  m_digitalWrite(TDIPIN,dw&0x80?HIGH:LOW); if(m_digitalRead(TDOPIN)==HIGH) dr|=0x80; tck();  
  return dr;
}
inline byte aser_rw(byte dw){
  byte dr=0;
  m_digitalWrite(TDIPIN,dw&0x01?HIGH:LOW); if(m_digitalRead(ADOPIN)==HIGH) dr|=0x01; tck();
  m_digitalWrite(TDIPIN,dw&0x02?HIGH:LOW); if(m_digitalRead(ADOPIN)==HIGH) dr|=0x02; tck();
  m_digitalWrite(TDIPIN,dw&0x04?HIGH:LOW); if(m_digitalRead(ADOPIN)==HIGH) dr|=0x04; tck();
  m_digitalWrite(TDIPIN,dw&0x08?HIGH:LOW); if(m_digitalRead(ADOPIN)==HIGH) dr|=0x08; tck();
  m_digitalWrite(TDIPIN,dw&0x10?HIGH:LOW); if(m_digitalRead(ADOPIN)==HIGH) dr|=0x10; tck();
  m_digitalWrite(TDIPIN,dw&0x20?HIGH:LOW); if(m_digitalRead(ADOPIN)==HIGH) dr|=0x20; tck();
  m_digitalWrite(TDIPIN,dw&0x40?HIGH:LOW); if(m_digitalRead(ADOPIN)==HIGH) dr|=0x40; tck();
  m_digitalWrite(TDIPIN,dw&0x80?HIGH:LOW); if(m_digitalRead(ADOPIN)==HIGH) dr|=0x80; tck(); 
  return dr;
}
inline void pio_write(byte c){
  m_digitalWrite(TCKPIN,c&TCKMASK?HIGH:LOW);
  m_digitalWrite(TMSPIN,c&TMSMASK?HIGH:LOW);
  m_digitalWrite(nCEPIN,c&nCEMASK?HIGH:LOW);
  m_digitalWrite(nCSPIN,c&nCSMASK?HIGH:LOW);
  m_digitalWrite(TDIPIN,c&TDIMASK?HIGH:LOW);
  m_digitalWrite(LEDPIN,c&OPEMASK?HIGH:LOW);
}
inline byte pio_read(){
  byte ret = 0;
  if(m_digitalRead(TDOPIN)==HIGH) ret|=0x01;
  if(m_digitalRead(ADOPIN)==HIGH) ret|=0x02;
  return ret;
}

//FastSerial Functions
struct FIFO{
  byte buffer[256];
  volatile byte rp;
  volatile byte wp;
} tx_fifo, rx_fifo;
SIGNAL(USART_RX_vect){
  byte c = UDR0;
  byte i = rx_fifo.wp + 1;
  if(i!=rx_fifo.rp) rx_fifo.buffer[rx_fifo.wp++] = c;
}
ISR(USART_UDRE_vect){
  if(tx_fifo.rp==tx_fifo.wp) bitClear(UCSR0B, UDRIE0);
  else UDR0 = tx_fifo.buffer[tx_fifo.rp++];
}
void Serial_begin(unsigned long baud){
  uint16_t baud_setting;
  rx_fifo.wp = 0; rx_fifo.rp = 0;
  tx_fifo.wp = 0; tx_fifo.rp = 0;
  UCSR0A = 1 << U2X0;
  baud_setting = (F_CPU / 4 / baud - 1) / 2;
  UBRR0H = baud_setting >> 8;
  UBRR0L = baud_setting;
  bitSet(UCSR0B, RXEN0);
  bitSet(UCSR0B, TXEN0);
  bitSet(UCSR0B, RXCIE0);
  bitClear(UCSR0B, UDRIE0);
}
inline byte Serial_available(){
  return rx_fifo.wp-rx_fifo.rp;
}
inline byte Serial_read(){
  byte c = rx_fifo.buffer[rx_fifo.rp];
  if(Serial_available()) rx_fifo.rp++;
  return c;
}
inline void Serial_write(byte c){
  byte i = tx_fifo.wp + 1;
  while(i==tx_fifo.rp){};
  tx_fifo.buffer[tx_fifo.wp++] = c;
  bitSet(UCSR0B, UDRIE0);
}

